import os
import tomli

"""
This File takes the prepared Data structures and writes the OPC-UA Server based on them.
    Methods: defFunction, getVariantType, writebeginning, writeenums, openobject, writemeasurement, writeparameter,
             writefunction, writeobjects, writeinterface, writeevents, writestreams, writendpt1, triggerevents, triggerstreams,
             writendpt2
"""


def writebeginning():
    """
    delets the existing opcua-server.py and generates a new one. This new file is started with all Imports and the needed set up
    is written. The objecttypes measurement and parameter are defined here too
    arguments: None
    returns: None
    """
    if os.path.exists("opcua-server.py"):
        os.remove("opcua-server.py")
    server = open("opcua-server.py", "a")
    server.write("""import logging 
import asyncio

from asyncua import ua, Server
from asyncua.common.structures104 import new_enum

logging.basicConfig(level=logging.INFO)
_logger = logging.getLogger('asyncua')
""")

    server.write("\nasync def main():")
    server.write(f"""
    # setup our server
    server = Server()
    await server.init()
    server.set_endpoint('{server_endpoint}')

    # setup our own namespace, not really necessary but should as spec
    uri = '{namespace}'
    idx = await server.register_namespace(uri)
    #id of rangetype
    rangeid = await server.nodes.base_data_type.get_child(["0:Structure", "0:Range"])



    #creating dataypes for parameters and measurements
    measurementtype = await server.nodes.base_data_type.add_data_type(idx, "Measurement")
    parametertype = await server.nodes.base_data_type.add_data_type(idx, "Parameter")
    
""")


def getVariantType(datatype):
    """
    returns the ua.VariantType according to the datatype of a variable.
        arguments:
            - the datatype
        returns:
            - ua.VarianType.<datatype> as String
                (float and double get casted to ua.VariantType.Double and all forms of int get casted to ua.VariantType.Int64)
    """
    varianttype = "ua.VariantType.Int64"
    if datatype == "float" or datatype == "double":
        varianttype = "ua.VariantType.Double"
    elif datatype == "boolean":
        varianttype = "ua.VariantType.Boolean"
    elif datatype == "time":
        varianttype = "ua.VariantType.DateTime"
    elif datatype == "string":
        varianttype = "ua.VariantType.String"
    elif datatype == "enum":
        varianttype = "ua.VariantType.String"
    return varianttype


with open("config.toml", "rb") as t:
    """
    opens toml config file to access its data
    """
    toml_dict = tomli.load(t)

#sets opcua-server address and namespace according to config file 
server_endpoint = toml_dict['server']
namespace = toml_dict['namespace']


def writeenums(enumslist):
    """
    Creates the needed enums in the beginning of the server. And creates three dicts needed for Information.
    :param enumslist:
    :return enumnodenames: Dict with format: {uuid of enum: opc ua object node of the enum}
    :return returnvalues: Dict with format: {uuid of enum: valuelist with
                            added numbers to values showing their corresponding int Value}
    """
    server = open("opcua-server.py", "a")
    writeindent = "    "
    returnvalues = {}
    enumnodenames = {}
    for enum in enumslist:
        returnvallist = []
        enumname = (enum["name"].lower()).replace(" ", "")
        values = enum["values"]
        for i in range(len(values)):
            retval = str(i) + ":" + str(values[i])
            returnvallist.append(retval)
            server.write(writeindent + "#creating the enums\n")
        server.write(writeindent + enumname + "= await new_enum(server, idx, \""+ str(enum["name"]) + "\", values=" + str(enum["values"]) + ")\n\n")
        enumnodenames[enum["uuid"]] = enumname
        returnvalues[enum["uuid"]] = returnvallist
    return enumnodenames, returnvalues


def openobject(objname, description):
    """
    Creates the needed objecttype
    :param objname: Name of object
    :param description:Description of the object
    :return objtypename: the name of the new node
    """
    server = open("opcua-server.py", "a")
    writeindent = "    "
    server.write(writeindent + "#creating the objecttype\n")
    objtypename = (objname.lower()).replace(" ", "") + "type"
    server.write(writeindent + objtypename + "= await server.nodes.base_object_type.add_object_type(idx, \"" + objname + "\" )\n")
    server.write(writeindent + objtypename + "description = await " + objtypename + ".add_property(idx, \"Description\", \"" + description + "\")\n")
    server.write(writeindent + "await " + objtypename + "description.set_modelling_rule(True)\n\n")
    return objtypename


def writemeasurement(measurementobj, headnode, enumnodedict, enumvaldict, varlist):
    """
    Creates the measurement variables with their properties
    :param measurementobj: a list of measurementobjects
    :param headnode: the component to which the varianles belong
    :param enumnodedict: Dict for the names of the enum nodes from writeenums (used for writeenumobj)
    :param enumvaldict: Dict for the enumvalues from writenums (used for writeenumobj)
    :param varlist: List of all the in measurements of the componengt in Format {name : value}
    :return: None
    """
    server = open("opcua-server.py", "a")
    writeindent = "    "
    measurementdict = {}
    if len(measurementobj) > 0:
        server.write(writeindent + "# measurements for " + headnode + "\n")
    for i in measurementobj:
        varname = ""
        measname = ""
        for fobj in varlist:
            if fobj["value"] == i.uuid:
                measname = fobj["name"]
                varname = (fobj["name"].lower()).replace(" ", "")
                varlist.remove(fobj)
                break
        if i.datatype == "enum":
            measurementdict[i.name] = varname
            server.write(writeindent + varname + " = await " + headnode + ".add_variable(idx,  \"" + measname + "\", " + i.value + ", datatype=" + enumnodedict[i.uuid] + ".nodeid)\n")
            server.write(writeindent + "await " + varname + "." + "set_modelling_rule(True)\n")
            count = 0
            for value in enumvaldict[i.uuid]:
                valname = "value" + str(count)
                server.write(writeindent + valname + " = await " + varname + ".add_property(idx, \"" + str(value) + "\", \"" + str(value) + "\")\n")
                server.write(writeindent + "await " + valname + "." + "set_modelling_rule(True)\n")
                i += 1
        else:
            varianttype = getVariantType(i.datatype)
            datatype = "measurementtype.nodeid"
            server.write(writeindent + varname + " = " + "await " + headnode + ".add_variable(idx, \"" + str(
               measname) + "\", " + str(i.value) + ", " + varianttype + ", " + datatype + ")\n")
            server.write(writeindent + "await " + varname + "." + "set_modelling_rule(True)\n")
            if i.range is not None and i.range != [None, None]:
                rangename = "rangeprop"
                server.write(writeindent + rangename + " = await " + varname + ".add_property(idx, \"range\", ua.Range(" + str(i.range[0]) + ", " + str(i.range[1]) + "))\n")
                server.write(writeindent + "await " + rangename + ".set_modelling_rule(True)\n")
            if i.unit != "UNITLESS":
                unitname = "unitprop"
                server.write(
                    writeindent + unitname + " = await " + varname + ".add_property(idx, \"unit\"," + "\"" + i.unit + "\")\n")
                server.write(writeindent + "await " + unitname + ".set_modelling_rule(True)\n")
        server.write(
            writeindent + varname + "description = await " + varname + ".add_property(idx, \"Description\", " + "\"" + i.description + "\")" + "\n")
        server.write(writeindent + "await " + varname + "description.set_modelling_rule(True)\n\n")


def writeparameter(parameterobj, headnode, enumnodedict, enumvaldict, parameterlist):
    """
    Creates the parameter values. (Similar to writemeasurement)
    :param parameterobj: A list of the objparameters
    :param headnode: The component to which the variables belong
    :param enumnodedict: Dict for the names of the enum nodes from writeenums (used for writeenumobj)
    :param enumvaldict: Dict for the enumvalues from writenums (used for writeenumobj)
    :param parameterlist: List of all the parameters of the component with Format {name: value}
    :return: None
    """
    writeindent = "    "
    server = open("opcua-server.py", "a")
    parameterdict = {}
    if len(parameterobj) > 0:
        server.write(writeindent + "# parameters for " + headnode + "\n")
    for i in parameterobj:
        varname = ""
        parname = ""
        for fobj in parameterlist:
            if fobj["value"] == i.uuid:
                parname = fobj["name"]
                varname = (fobj["name"].lower()).replace(" ", "")
                parameterlist.remove(fobj)
                break
        if i.datatype == "enum":
            server.write(writeindent + varname + " = await " + headnode + ".add_variable(idx,  \"" + parname + "\", " + i.value + ", datatype=" + enumnodedict[i.uuid] + ".nodeid)\n")
            server.write(writeindent + "await " + varname + "." + "set_modelling_rule(True)\n")
            count = 0
            for value in enumvaldict[i.uuid]:
                valname = "value" + str(count)
                server.write(writeindent + valname + " = await " + varname + ".add_property(idx, \"" + str(value) + "\", \"" + str(value) + "\")\n")
                server.write(writeindent + "await " + valname + "." + "set_modelling_rule(True)\n")
                count += 1
        else:
            varianttype = getVariantType(i.datatype)
            datatype = "parametertype.nodeid"
            parameterdict[i.name] = varname
            server.write(writeindent + varname + " = " + "await " + headnode + ".add_variable(idx, \"" + str(
                parname) + "\", " + str(i.value) + ", " + varianttype + ", " + datatype + ")\n")
            server.write(writeindent + "await " + varname + "." + "set_modelling_rule(True)\n")
            if i.range is not None and i.range != [None, None]:
                rangename = "rangeprop"
                server.write(
                    writeindent + rangename + " = await " + varname + ".add_property(idx, \"range\", ua.Range(" + str(i.range[0]) + ", " + str(i.range[1]) + "))\n")
                server.write(writeindent + "await " + rangename + ".set_modelling_rule(True)\n")
            if i.unit != "UNITLESS":
                unitname = "unitprop"
                server.write(
                    writeindent + unitname + " = await " + varname + ".add_property(idx, \"unit\"," + "\"" + i.unit + "\")\n")
                server.write(writeindent + "await " + unitname + ".set_modelling_rule(True)\n")

        server.write(
            writeindent + varname + "description = await " + varname + ".add_property(idx, \"Description\", " + "\"" + i.description + "\")" + "\n")
        server.write(writeindent + "await " + varname + "description.set_modelling_rule(True)\n\n")


def defFunction(functionname, file, indent, inargs, function, returnargs):
    #return values are not written in the function at the moment
    """
    writes the body for each function, it leaves the body empty and gives a description for the method which should be implemented
    :param functionname: name of the function changed to a lowercase String without blanks
    :param file: file it writes in
    :param indent: indent it writes with
    :param inargs: arguments of the function
    :param function: the object
    :param returnargs: arguments that should be returned to be listed in the comments
    """
    writeindent = ""
    for i in range(indent):
        writeindent = writeindent + "    "
    file.write(writeindent + "async def " + functionname + "(parent")
    if inargs is not None:
        file.write(", ")
        for i in range(len(inargs)):
            if i < (len(inargs) - 1):
                file.write(inargs[i] + ", ")
            else:
                file.write(inargs[i])
    file.write("):\n")
    writeindent = writeindent + "    "
    file.write(writeindent + "#ToDo: Write a Function that fits the decription: \"" + str(function.description) + "\"\n")
    file.write(writeindent + "    # Use \"await\" before changing values and cast to Ua Varianttype\n")
    file.write(writeindent + "    # Use get and set_value to change the value at a given nodeid e.g: await testvar.set_value(ua.Variant(position))\n")
    if returnargs is not None:
        file.write(writeindent + "    # Function should return these values: " + str(returnargs) + "\n")
    file.write(writeindent + "pass\n")


def writefunction(functionobj, headnode, functionlist):
    """
    Takes the function objects writes the function body by calling deffunction and adds it to the objecttypes
    :param functionobj: List of functionobjects
    :param headnode: component the functions belong to
    :param functionlist: List of all functions from the component with the Format {name: value}
    :return: None
    """
    server = open("opcua-server.py", "a")
    writeindent = "    "
    functionname = ""
    funcname = ""
    for i in functionobj:
        for fobj in functionlist:
            if fobj["value"] == i.uuid:
                funcname = fobj["name"]
                functionname = (fobj["name"].lower()).replace(" ", "")
                functionlist.remove(fobj)
                break
        defFunction(functionname, server, 1, i.inargsnames, i, i.outargsnames)
        server.write("\n")
        inargslist = []
        if i.inargsdatatypes != None:
            for j in i.inargsdatatypes:
                inargslist.append(getVariantType(j))
        outargslist = []
        if i.outargsdatatypes != None:
            for j in i.outargsdataypes:
                inargslist.append(getVariantType(j))
        methodvarname = functionname + "var"
        server.write(writeindent + methodvarname + " = await " + headnode + ".add_method(idx, \"" + funcname +"\", " + functionname + ", [")
        for a in range(len(inargslist)):
            server.write(str(inargslist[a].replace("\'", "")))
            if a < len(inargslist) - 1:
                server.write(", ")
        server.write("], [")
        for a in range(len(outargslist)):
            server.write(str(outargslist[a].replace("\'", "")))
            if a < len(outargslist) - 1:
                server.write(", ")
        server.write("]")
        server.write(")\n")
        server.write(writeindent + "await " + methodvarname + ".set_modelling_rule(True)\n")
        server.write(writeindent + methodvarname + "description = await " + methodvarname + ".add_property(idx, \"Description\", \"" + i.description + "\")\n")
        server.write(writeindent + "await " + methodvarname + "description.set_modelling_rule(True)\n\n")


def writeobjects(objtype, objname, headnode):
    """
    Generates the objects of the components.
    :param objtype: The node of the node with the objecttype
    :param objname: The name of the object
    :param headnode: The node the component belongs to.
    :return: The name of the newly created objectnode
    """
    server = open("opcua-server.py", "a")
    writeindent = "    "

    server.write(writeindent + "# creating object \n")
    objtypename = (objname.lower()).replace(" ", "") + "var"
    server.write(
        writeindent + objtypename + "= await " + headnode + ".add_object(idx, \"" + objname + "\", " + objtype + ".nodeid)\n\n")
    return objtypename


def writeinterface(objtype, objname):
    """
    Generates the interface, meaning the highest object layer. (Works similar to writeobjects)
    :param objtype: The name of the node with the objecttype
    :param objname: The name of the object
    :return: The name of the newly created objectnode
    """
    server = open("opcua-server.py", "a")
    writeindent = "    "
    server.write(writeindent + "# creating Interface\n")
    objtypename = (objname.lower()).replace(" ","") + "var"
    server.write(writeindent + objtypename + "= await server.nodes.objects.add_object(idx, \"" + objname + "\", " + objtype + ".nodeid)\n\n")
    return objtypename


def writeevents(events, component, eventlist):
    """
    Generates the events and event generator for each event and adds the fitting severity.
    :param events: List of eventdicts
    :param component: The component the events belong to
    :param eventlist: Empty dict (filled with the names of the eventgenerator nodes)
    :return: The eventlist now filled in the format {event name(generated in the code) : eventgenerator node name}
    """
    server = open("opcua-server.py", "a")
    writeindent = "    "
    if len(events) > 0:
        server.write(writeindent + "# creating event and eventgenerators for " + component + "\n")
    for event in events:
        eventname = (event["element"].lower()).replace(" ", "") + component + str(event["severity"])
        eventgenname = eventname + "generator"
        eventlist[eventname] = eventgenname
        severity = 0
        if event["severity"] == "debug":
            severity = 200
        elif event["severity"] == "info":
            severity = 400
        elif event["severity"] == "warning":
            severity = 600
        elif event["severity"] == "error":
            severity = 800
        elif event["severity"] == "critical":
            severity = 1000
        server.write(writeindent + eventname + " = await server.create_custom_event_type(idx, \"" + str(event["element"]) + " " + str(event["severity"]) + "\", ua.ObjectIds.SystemEventType)\n")
        server.write(writeindent + str(eventlist[eventname]) + " = await server.get_event_generator(" + eventname + ", " + component + ")\n")
        server.write(writeindent + str(eventlist[eventname]) + ".event.Severity = " + str(severity) + "\n\n")
    return eventlist


def writestreams(streams, component, streamslist, variablelist, jsonmeasurements, jsonparameters):
    """
    Generates the streams as events withe event generator, uses standard severity 1.
    :param streams: List of streamdicts
    :param component: Component the streams belong to.
    :param streamslist: Empty dict (filled with streamgenerator node names later on)
    :param variablelist: List of the variables belonging to the component in Format {name : value}
    :param jsonmeasurements: List of all measurement elements
    :param jsonparameters: List of all parameter elements
    :return streamlist: The streamslist now filled in the format
                            {streamname(generated in the code) : streamgenerator node name}
    :return updatestreamsdict: A dict in which the compare variable names for
                                update streams are saved in format {streamname: variablename}
    """
    server = open("opcua-server.py", "a")
    writeindent = "    "
    updatestreamdict = {}
    if len(streamslist) > 0:
        server.write(writeindent + "# creating events and event generators for" + component + "\n")
    for stream in streams:
        streamname = (stream["measurement"].lower()).replace(" ", "") + component
        streamgenname = streamname + "generator"
        streamslist[streamname] = streamgenname
        server.write(writeindent + streamname + " = await server.create_custom_event_type(idx, \"" + str(stream["measurement"]) + "\", ua.ObjectIds.SystemEventType)\n")
        server.write(writeindent + str(streamslist[streamname]) + " = await server.get_event_generator(" + streamname + ", " + component + ")\n")

        if stream["streamType"] == "update":
            value = 0
            for variable in variablelist:
                if variable["name"] == stream["measurement"]:
                    for measurement in jsonmeasurements:
                        if measurement["uuid"] == variable["value"]:
                            value = measurement["value"]
                    for parameter in jsonparameters:
                        if parameter["uuid"] == variable["value"]:
                            value = parameter["value"]

            savevaluename = "save" + streamname
            writevalue = ""
            if type(value) == str:
                writevalue = "\"" + value + "\""
            else:
                writevalue = str(value)
            server.write(writeindent + savevaluename + " = " + writevalue + "\n\n")
            updatestreamdict[streamname] = savevaluename

    return streamslist, updatestreamdict


def writendpt1():
    """
    Writes the beginning of the while loop which starts the server.
    :return: None
    """
    server = open("opcua-server.py", "a")
    server.write("""
    async with server:
        print("Server is running!")
        count = 0
        
        while True:\n""")


def triggerevents(generatordict, eventslist, component, variablelist):
    """
    Writes the if cases in which the events get triggered.
    :param generatordict: Dict with the format
                            {eventname (generated according to format from writeevents) : eventgeneratorname}
    :param eventslist: List of eventdicts
    :param component: Component the events belong to
    :param variablelist: List of all variabledicts from component in the format {name : value}
    :return: None
    """
    server = open("opcua-server.py", "a")
    indent = 3
    variablenamelist = {}
    if len(eventslist) > 0:
        server.write("          # triggering events from " + component + "\n")
    for event in eventslist:
        writeindent = ""
        for i in range(indent):
            writeindent += "    "
        valuename = event["element"] + component
        eventsearchname = ""

        if event["element"] in list(variablenamelist.keys()):
            eventsearchname = variablenamelist[str(event["element"])]
        else:
            for variable in variablelist:
                if variable["name"] == event["element"]:
                    eventsearchname = variable["name"]
            variablenamelist[str(event["element"])] = eventsearchname
            server.write(writeindent + valuename + "= await " + component + ".get_child([\"2:" + eventsearchname + "\"])\n")
        dictname = (event["element"].lower()).replace(" ", "") + component + event["severity"]
        generatorname = generatordict[dictname]

        trigger = str(event["target"])
        if type(event["target"]) == str:
            trigger = "\"" + str(event["target"] + "\"")
        server.write(writeindent + "if (await " + valuename + ".get_value()) " + event["trigger"] + " " + trigger + ":\n")
        writeindent += "    "
        server.write(writeindent + "await " + generatorname + ".trigger(message=\"" + event["message"] + "\")\n\n")
        writeindent = ""
        for i in range(indent):
            writeindent += "    "


def triggerstreams(generatordict, streamslist, component, variablelist, savevaldict):
    """
    Generates the Streams according to their stream type. (Dynamic streams are still missing)
    :param generatordict: Dict with the format
                            {streamname (generated according to format from writestreams) : streamgeneratorname}
    :param streamslist: List of streamdicts
    :param component: Component the streams belong to
    :param variablelist: List of all variabledicts from the component in the format {name: value}
    :param savevaldict: Dict with the name of the compare values for update streams.
                            Format {streamname : comparevaluename}
    :return: None
    """
    server = open("opcua-server.py", "a")
    indent = 3
    if len(streamslist) > 0:
        server.write("          # triggering streams from " + component + "\n")
    for stream in streamslist:
        writeindent = ""
        streamname = ""
        valuename = "stream" + ((stream["measurement"]).lower()).replace(" ", "") + component
        for i in range(indent):
            writeindent += "    "
        for variable in variablelist:
            if variable["name"] == stream["measurement"]:
                streamname = variable["name"]
        server.write(writeindent + valuename + " = await " + component + ".get_child([\"2:" + streamname + "\"])\n")
        generatorname = generatordict[(stream["measurement"].lower()).replace(" ", "") + component]
        if stream["streamType"] == "update":
            server.write(writeindent + "if (await " + valuename + ".get_value()) != " + savevaldict[(stream["measurement"].lower()).replace(" ", "") + component] + ":\n")
            writeindent += "    "
            server.write(writeindent + "puffer = await " + valuename + ".get_value()\n")
            server.write(writeindent + "await " + generatorname + ".trigger(message= \"value\" + str(puffer))\n")
            server.write(writeindent + savevaldict[(stream["measurement"].lower()).replace(" ", "") + component] + " = puffer \n\n")
            writeindent = ""
            for i in range(indent):
                writeindent += "    "

        if stream["streamType"] == "fixed":
            server.write(writeindent + stream["measurement"] + component + "time = count % " + str(stream["intervalValue"]) + "\n")
            server.write(writeindent + "if "+ stream["measurement"] + component + "time == 0:\n")
            writeindent += "    "
            server.write(writeindent + "puffer = await " + valuename + ".get_value()\n")
            server.write(writeindent + "await " + generatorname + ".trigger(message=\"" + stream["measurement"] + " value: \" + str(puffer))\n\n")
            writeindent = ""
            for i in range(indent):
                writeindent += "    "


def writendpt2():
    """
    Writes the end of the server with the await asyncio.sleep(1) and the time counter.
    :return: None
    """
    server = open("opcua-server.py", "a")
    server.write("""
            await asyncio.sleep(1)
            count += 1

if __name__ == '__main__':
    asyncio.run(main())""")